/*
 * DxFlexServer.c - A AX.25 broadcasting DX cluster client for PC/Flexnet 32 bit - version 1.0
 *
 * Copyright (C) 2011 Marius Petrescu, YO2LOJ, <marius@yo2loj.ro>.
 *
 * This software acts as a AX.25 monitor towards Flexnet and as a telnet server on the localhost
 * port 23 (telnet). Heard DX cluster announcements will be put out on the telnet server.
 *
 * Compile with Open Watcom C
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * Version 1.0 - 5. Dec. 2011
 *
 */

#include <windows.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <winsock.h>
#include "flexappl.h"
#include "dxflex.h"


#define PROGRAM_NAME    "DxFlexDaemon"
#define WINSOCK_VER     0x0101


HINSTANCE       MyInstance;
static char    DxFlexClass[32]="DxFlexClass";

char *szServerHello = "\r*** YO2LOJ DX Cluster Flexnet Client ***\n\n\r";
char *szServerPrompt = "DXBC-0>";

HWND        hwnd;
WSADATA     wsadata;
TRACE       trace;

TIMERPROC lpfnDaemonTimerProc;
#define ID_DAEMONTIMER  0x1000

static BOOL FirstInstance( HINSTANCE );
static BOOL AnyInstance( HINSTANCE, int, LPSTR );

typedef struct t_sockent {
    SOCKET s;
    struct t_sockent *next;
} sockent;

sockent se_base = {0, NULL};

void sockent_add(SOCKET s)
{
    sockent *se = &se_base;
    sockent *ns;
    
    while(se->next !=NULL) {
        se = se->next;
    }
    ns = malloc(sizeof(sockent));
    if (ns == NULL) {
        MessageBox(NULL, "malloc() failed", PROGRAM_NAME, MB_OK|MB_ICONSTOP);
    }
    else {
        se->next = ns;
        ns->s = s;
        ns->next = NULL;
    }
}

void sockent_del(SOCKET s)
{
    sockent *se = &se_base;
    sockent *rem;
    while (se->next != NULL) {
        if (se->next->s == s) {
            rem = se->next;
            se->next = se->next->next;
            free(rem);
            break;
        }
        se = se->next;
    }
}

int PASCAL DaemonTimerProc(void)
{
        return 0;
}

long _export FAR PASCAL WindowProc( HWND, unsigned, UINT, LONG );

/*
 * WinMain - initialization, message loop
 */
int PASCAL WinMain( HINSTANCE this_inst, HINSTANCE prev_inst, LPSTR cmdline,
                    int cmdshow )
{
#ifdef __NT__
    typedef VOID (WINAPI *PFNICC)( VOID );
    HINSTANCE   commctrl_inst;
    PFNICC      icc;
#endif
    MSG         msg;

#ifdef __NT__
    commctrl_inst = LoadLibrary( "comctl32.dll" );
    if( commctrl_inst != NULL ) {
        icc = (PFNICC)GetProcAddress( commctrl_inst, "InitCommonControls" );
        if( icc != NULL ) {
            icc();
        }
    }
#endif

    MyInstance = this_inst;
#ifdef __WINDOWS_386__
    sprintf( GenericClass, "GenericClass%d", this_inst );
    prev_inst = 0;
#endif
    if( !prev_inst ) {
        if( !FirstInstance( this_inst ) ) return( FALSE );
    }
    if( !AnyInstance( this_inst, cmdshow, cmdline ) ) return( FALSE );

    while( GetMessage( &msg, NULL, 0, 0 ) ) {

        TranslateMessage( &msg );
        DispatchMessage( &msg );

    }

#ifdef __NT__
    if( commctrl_inst != NULL ) {
        FreeLibrary( commctrl_inst );
    }
#endif

    ax_exit();
    WSACleanup();

    return( msg.wParam );

} /* WinMain */

/*
 * FirstInstance - register window class for the application,
 *                 and do any other application initialization
 */
static BOOL FirstInstance( HINSTANCE this_inst )
{
    WNDCLASS    wc;
    BOOL        rc;

    /*
     * set up and register window class
     */
    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = (LPVOID) WindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = sizeof( DWORD );
    wc.hInstance = this_inst;
    wc.hIcon = LoadIcon( this_inst, "DxFlexIcon" );
    wc.hCursor = LoadCursor( NULL, IDC_ARROW );
    wc.hbrBackground = GetStockObject( WHITE_BRUSH );
    wc.lpszMenuName = "DxFlexMenu";
    wc.lpszClassName = DxFlexClass;
    rc = RegisterClass( &wc );
    return( rc );

} /* FirstInstance */

/*
 * AnyInstance - do work required for every instance of the application:
 *                create the window, initialize data
 */
static BOOL AnyInstance( HINSTANCE this_inst, int cmdshow, LPSTR cmdline )
{

    SOCKET      s = INVALID_SOCKET;
    LPSERVENT   lpservent;
    short      TelnetPort;
    SOCKADDR_IN sock_addr;

    /*
     * create main window
     */
    hwnd = CreateWindow(
        DxFlexClass,            /* class */
        "Dx Flexnet Client",    /* caption */
        WS_OVERLAPPEDWINDOW,    /* style */
        CW_USEDEFAULT,          /* init. x pos */
        CW_USEDEFAULT,          /* init. y pos */
        320,                    /* init. x size */
        240,                    /* init. y size */
        NULL,                   /* parent window */
        NULL,                   /* menu handle */
        this_inst,              /* program handle */
        NULL                    /* create parms */
        );

    if( !hwnd ) return( FALSE );

    /*
     * display window
     */
    ShowWindow( hwnd, cmdshow );
    UpdateWindow( hwnd );

    if (!SetTimer(hwnd, ID_DAEMONTIMER, 100, NULL)) {
        MessageBox(NULL, "No timers available.", PROGRAM_NAME, MB_OK|MB_ICONSTOP);
        return FALSE;
    }
    if (WSAStartup(WINSOCK_VER, &wsadata)) {
        MessageBox(NULL, "Error loading WINSOCK.DLL.", PROGRAM_NAME, MB_OK|MB_ICONSTOP);
        WSACleanup();
        return FALSE;
    }
    if (wsadata.wVersion < WINSOCK_VER) {
        MessageBox(NULL, "Incorrect WINSOCK.DLL version.", PROGRAM_NAME, MB_OK|MB_ICONSTOP);
        WSACleanup();
        return FALSE;
    }
    if (wsadata.iMaxSockets < 5) {
        MessageBox(NULL, "Not enough free sockets.", PROGRAM_NAME, MB_OK|MB_ICONSTOP);
        WSACleanup();
        return FALSE;
    }

    if (ax_init() != 1) {
        MessageBox(NULL, "Flexnet init failed.", PROGRAM_NAME, MB_OK|MB_ICONSTOP);
        return FALSE;
    }

    trace. ch_mask = 0xFF;  // monitor all channels
    trace. typfilter = TRACETYP_NO_S;
    trace. trxfilter = 1; // only Rx
    trace.call[0] = 0; // listen to all calls
    // start trace
    if (!l2_set_monitor(&trace)) {
        MessageBox(NULL, "Flexnet start trace failed.", PROGRAM_NAME, MB_OK|MB_ICONSTOP);
        return FALSE;       
    }
    
    lpservent = getservbyname("telnet", "tcp");
    if (lpservent == NULL) TelnetPort = IPPORT_TELNET;
    else TelnetPort = lpservent->s_port;

    s = socket(AF_INET, SOCK_STREAM, 0);
    if (s == INVALID_SOCKET) {
        MessageBox(NULL, "Cannot create server socket.", PROGRAM_NAME, MB_OK|MB_ICONSTOP);
        SendMessage(hwnd, WM_DESTROY, 0, 0);
        return 0;
    }
    sock_addr.sin_family = AF_INET;
    sock_addr.sin_port = TelnetPort;
    sock_addr.sin_addr.s_addr = INADDR_ANY;
    
    if (bind(s, (struct sockaddr FAR *)&sock_addr, sizeof(sock_addr)) == SOCKET_ERROR){
        MessageBox(NULL, "Cannot bind server socket.", PROGRAM_NAME, MB_OK|MB_ICONSTOP);
        SendMessage(hwnd, WM_DESTROY, 0, 0);
        return 0;
    }      
    if (listen(s, 2) == SOCKET_ERROR) {
        MessageBox(NULL, "Cannot listen for clients.", PROGRAM_NAME, MB_OK|MB_ICONSTOP);
        SendMessage(hwnd, WM_DESTROY, 0, 0);
        return 0;
    }
    if (WSAAsyncSelect(s, hwnd, WM_SOCKET, FD_ACCEPT|FD_CLOSE) == SOCKET_ERROR) {
        MessageBox(NULL, "WSAAsyncSelect did not work.", PROGRAM_NAME, MB_OK|MB_ICONSTOP);
        SendMessage(hwnd, WM_DESTROY, 0, 0);
        return 0;
    }
  
    return( TRUE );
} /* AnyInstance */

/*
 * AboutDlgProc - processes messages for the about dialog.
 */
BOOL _export FAR PASCAL AboutDlgProc( HWND hwnd, unsigned msg,
                                UINT wparam, LONG lparam )
{
    lparam = lparam;                    /* turn off warning */

    switch( msg ) {
    case WM_INITDIALOG:
        return( TRUE );

    case WM_COMMAND:
        if( LOWORD( wparam ) == IDOK ) {
            EndDialog( hwnd, TRUE );
            return( TRUE );
        }
        break;
    }
    return( FALSE );

} /* AboutDlgProc */

/*
 * WindowProc - handle messages for the main application window
 */
LONG _export FAR PASCAL WindowProc( HWND hwnd, unsigned msg,
                                     UINT wparam, LONG lparam )
{
    FARPROC         proc;
    SOCKET          s;
    char            buff[1024];
    int             len;
    sockent         *se = &se_base;
    const FRAME     *frame;
    
    switch( msg ) {
    case WM_COMMAND:
        switch( LOWORD( wparam ) ) {
        case MENU_EXIT:
            SendMessage(hwnd, WM_DESTROY, 0, 0);
            break;
        case MENU_ABOUT:
            proc = MakeProcInstance( (FARPROC)AboutDlgProc, MyInstance );
            DialogBox( MyInstance,"AboutBox", hwnd, (DLGPROC)proc );
            FreeProcInstance( proc );
            break;
        }
        break;

    case WM_SOCKET:
        if (WSAGETSELECTERROR(lparam))
        {
            closesocket((SOCKET)wparam);
            break;
        }
        switch (WSAGETSELECTEVENT(lparam))
        {
            case FD_ACCEPT:
                s = accept(wparam, NULL, NULL);
                if (s == INVALID_SOCKET) {
                    MessageBox(NULL, "Cannot create new client socket.", PROGRAM_NAME, MB_OK|MB_ICONSTOP);
                    SendMessage(hwnd, WM_DESTROY, 0, 0);
                    break;
                }
                               
                if (WSAAsyncSelect(s, hwnd, WM_SOCKET, FD_READ|FD_WRITE|FD_CLOSE) == SOCKET_ERROR)
                {
                    MessageBox(NULL, "WSAAsyncSelect did not work.", PROGRAM_NAME, MB_OK|MB_ICONSTOP);
                    SendMessage(hwnd, WM_DESTROY, 0, 0);
                    break;
                }
                sockent_add(s);
                break;
            case FD_READ:
                len = recv((SOCKET)wparam, buff, sizeof(buff), 0);
                buff[len] = 0;
                if (strncmp(buff, "/q", 2)==0) {
                    sockent_del((SOCKET)wparam);
                    closesocket((SOCKET)wparam);
                }
                break;
            case FD_WRITE:
                send((SOCKET)wparam, szServerHello, strlen(szServerHello), 0);
                send((SOCKET)wparam, szServerPrompt, strlen(szServerPrompt), 0);
                break;                
            case FD_CLOSE:
                sockent_del((SOCKET)wparam);
                closesocket((SOCKET)wparam);
                break;
        }
        break;

    case WM_TIMER:
        frame = l2_get_monitor();
        if (frame) {
            if (frame->textlen < (sizeof(buff)-3)) {
                strncpy(buff, frame->text, frame->textlen);
                buff[frame->textlen] = '\r';
                buff[frame->textlen + 1] = '\n';
                buff[frame->textlen + 2] = 0;
                if (strncmp(buff, "DX de ", 6) == 0) {
                    // strange length correction to have the time on the same position as direct telnet - insert a space (why the heck is this?)
                    memcpy(&buff[0x46], &buff[0x45], sizeof(buff)-0x47);
                    buff[0x45] = 0x20;
                    while (se->next != NULL) {
                        send(se->next->s, buff, strlen(buff), 0);
                        se = se->next;
                    }
                }
            }
            l2_ack_monitor();
        }
        break;
        
    case WM_DESTROY:
        PostQuitMessage( 0 );
        break;

    default:
        return( DefWindowProc( hwnd, msg, wparam, lparam ) );
    }
    return( 0L );

} /* WindowProc */

